//
//  BleManager.m
//  epuck2-ble-monitor
//
//  Created by Stefano Morgani on 01.05.19.
//  Copyright (c) 2019 GCtronic. All rights reserved.
//

#import "BleManager.h"

@implementation BleManager

@synthesize BTManager;
@synthesize txCharacteristic;

+ (CBUUID *) uartServiceUUID
{
    return [CBUUID UUIDWithString:@"0000abf0-0000-1000-8000-00805f9b34fb"];
}

+ (CBUUID *) txCharacteristicUUID
{
    return [CBUUID UUIDWithString:@"0000abf1-0000-1000-8000-00805f9b34fb"];
}

+ (CBUUID *) rxCharacteristicUUID
{
    return [CBUUID UUIDWithString:@"0000abf2-0000-1000-8000-00805f9b34fb"];
}

+ (id)sharedInstance {
    static BleManager *_sharedInstance = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        _sharedInstance = [[self alloc] init];
    });
    return _sharedInstance;
}

- (id)init {
    if (self = [super init]) {
        
        BTManager = [[CBCentralManager alloc] initWithDelegate:self queue:nil options:@{CBCentralManagerOptionRestoreIdentifierKey: @"myCentralManagerIdentifier"}];
        
        peripherals = [NSMutableArray arrayWithCapacity:1];
        
        sendingData = false;
        
    }
    return self;
}

- (void)dealloc {
    // Should never be called, but just here for clarity really.
}

- (BOOL)bleIsOff {
    return (BTManager.state == CBCentralManagerStatePoweredOff);
}

- (void)disconnectPeripheral {
    [BTManager cancelPeripheralConnection:blePeripheral];
}

- (NSInteger)peripheralsCount {
    if(peripherals == nil) {
        return 0;
    }
    return [peripherals count];
}

- (void)setCurrentPeripheral:(NSInteger)index {
    PeripheralCell *per = [peripherals objectAtIndex:index];
    blePeripheral = per.peripheral;
}

- (void)connectToPeripheral {
    [BTManager connectPeripheral:blePeripheral options:nil];
}

- (NSMutableArray *)getPeripherals {
    return peripherals;
}

- (NSString *)getPeripheralName:(NSInteger)index {
    if(index > ([self peripheralsCount]-1)) {
        return nil;
    } else {
        PeripheralCell *pcell = [peripherals objectAtIndex:index];
        return [pcell.peripheral name];
    }
}

- (NSString *)getCurrentPeripheralName {
    return [blePeripheral name];
}

- (int)getPeripheralRssi:(NSInteger)index {
    if(index > ([self peripheralsCount]-1)) {
        return -1;
    } else {
        PeripheralCell *pcell = [peripherals objectAtIndex:index];
        return [pcell.rssi intValue];
    }
}

- (void) clearDiscoveredPeripherals {
    [peripherals removeAllObjects];
}

- (void)startScanning {
    NSLog(@"start scanning");
    [BTManager scanForPeripheralsWithServices:[NSArray arrayWithObject:self.class.uartServiceUUID] options: [NSDictionary dictionaryWithObject:[NSNumber numberWithBool:NO] forKey:CBCentralManagerScanOptionAllowDuplicatesKey]];
}

- (void)stopScanning {
    [BTManager stopScan];
}

- (void)freeCurrentPeripheral {
    if(blePeripheral) {
        [blePeripheral setDelegate:nil];
        blePeripheral = nil;
    }
}

- (void)centralManager:(CBCentralManager *)central willRestoreState:(NSDictionary *)state {
    NSLog(@"willRestoreState");
}

- (void)centralManagerDidUpdateState:(CBCentralManager *)central {
    
    NSLog(@"centralManagerDidUpdateState");
    
    if (central.state == CBCentralManagerStatePoweredOff){
        NSLog(@"BLE OFF");

        if(blePeripheral) {
            [BTManager cancelPeripheralConnection:blePeripheral];
            [blePeripheral setDelegate:nil];
            blePeripheral = nil;
        }
        
        [peripherals removeAllObjects];
        
        
        // "didDisconnectPeripheral" will not be called when the Bluetooth is turned off while a device is connected,
        // thus adjust the interface directly here by notifying the view controller.
        [[NSNotificationCenter defaultCenter] postNotificationName:@"bleOffNotification" object:self userInfo:nil];
        
    } else if (central.state == CBCentralManagerStatePoweredOn){
        NSLog(@"BLE ON");   // value is 5
        
    } else if (central.state == CBCentralManagerStateUnknown){
        NSLog(@"NOT RECOGNIZED");
        
    } else if(central.state == CBCentralManagerStateUnsupported){
        NSLog(@"BLE NOT SUPPORTED");
        
    }
    
}

/**
 Called when scanner finds device
 First checks if it exists, if not then insert new device
 */
- (void) centralManager:(CBCentralManager *)central didDiscoverPeripheral:(CBPeripheral *)peripheral advertisementData:(NSDictionary *)advertisementData RSSI:(NSNumber *)RSSI {
    
    BOOL (^test)(id obj, NSUInteger idx, BOOL *stop);
    test = ^ (id obj, NSUInteger idx, BOOL *stop) {
        if([[[obj peripheral] name] compare:peripheral.name] == NSOrderedSame)
            return YES;
        return NO;
    };
    
    NSLog(@"peripheralsCount = %d", [self peripheralsCount]);
    
    PeripheralCell* cell;
    NSUInteger t = [peripherals indexOfObjectPassingTest:test];
    if(t != NSNotFound) {
        NSLog(@"found device already present");
        cell = [peripherals objectAtIndex:t];
        cell.peripheral = peripheral;
        cell.rssi = RSSI;
        
        // Notify view controller.
        [[NSNotificationCenter defaultCenter] postNotificationName:@"discoverUpdateBleDeviceNotification" object:self userInfo:@{@"index":[NSNumber numberWithInt:(int)t]}];
    } else {
        NSLog(@"found new device");
        cell=[[PeripheralCell alloc] init];
        [peripherals addObject: cell];
        cell.peripheral = peripheral;
        cell.rssi = RSSI;
        
        // Notify view controller.
        [[NSNotificationCenter defaultCenter] postNotificationName:@"discoverAddBleDeviceNotification" object:self userInfo:nil];
        
    }
    
}

/*
 connected to peripheral
 Show service search view
 */
- (void)centralManager:(CBCentralManager *)central didConnectPeripheral:(CBPeripheral *)peripheral {

    // Notify controller view.
    [[NSNotificationCenter defaultCenter] postNotificationName:@"connectedBleDeviceNotification" object:self userInfo:nil];
    
    sendingData = false;

    [peripheral setDelegate:self];
    [peripheral discoverServices:nil];
    
}

- (void)centralManager:(CBCentralManager *)central didDisconnectPeripheral:(CBPeripheral *)peripheral error:(NSError *)error {
    NSLog(@"Peripheral diconnected!");
    
    sendingData = false;
        
    // Notify controller view and let it decide whether to reconnect or not.
    [[NSNotificationCenter defaultCenter] postNotificationName:@"disconnectedBleDeviceNotification" object:self userInfo:nil];
}

/*
 Invoked whenever the central manager fails to create a connection with the peripheral.
 */
- (void)centralManager:(CBCentralManager *)central didFailToConnectPeripheral:(CBPeripheral *)aPeripheral error:(NSError *)error {
    
    NSLog(@"Fail to connect to peripheral: %@ with error = %@", aPeripheral, [error localizedDescription]);
    
    
    // Notify controller view and let it decide whether to reconnect or not.
    [[NSNotificationCenter defaultCenter] postNotificationName:@"failToConnectBleDeviceNotification" object:self userInfo:nil];
    
}

- (void)peripheral:(CBPeripheral *)peripheral didDiscoverServices:(NSError *)error {
    
    NSLog(@"didDiscoverServices");
    
    for (CBService *aService in peripheral.services)
    {
        NSLog(@"Service found with UUID: %@", aService.UUID);
        
        /* Uart service (raw sensors) */
        if ([aService.UUID isEqual:self.class.uartServiceUUID])
        {
            //NSLog(@"Discovering characteristics");
            [peripheral discoverCharacteristics:@[self.class.txCharacteristicUUID, self.class.rxCharacteristicUUID] forService:aService];
        }
        
    }
    
}

- (void)peripheral:(CBPeripheral *)peripheral didDiscoverCharacteristicsForService:(CBService *)service error:(NSError *)error
{
    
    NSLog(@"didDiscoverCharacteristicsForService with UUID: %@", service.UUID);
    
    if(error != nil)
    {
        NSLog(@"%@", [error localizedDescription]);
        //TODO: handle error
        return;
    }
    
    if ([service.UUID isEqual:[CBUUID UUIDWithString:@"180D"]]) // Heart rate service.
    {
        for (CBCharacteristic *aChar in service.characteristics)
        {
            /* Set notification on heart rate measurement */
            if ([aChar.UUID isEqual:[CBUUID UUIDWithString:@"2A37"]])
            {
                [peripheral setNotifyValue:YES forCharacteristic:aChar];
                NSLog(@"Found a Heart Rate Measurement Characteristic");
            }
            /* Read body sensor location */
            if ([aChar.UUID isEqual:[CBUUID UUIDWithString:@"2A38"]])
            {
                [peripheral readValueForCharacteristic:aChar];
                NSLog(@"Found a Body Sensor Location Characteristic");
            }
            
            /* Write heart rate control point */
            if ([aChar.UUID isEqual:[CBUUID UUIDWithString:@"2A39"]])
            {
                uint8_t val = 1;
                NSData* valData = [NSData dataWithBytes:(void*)&val length:sizeof(val)];
                [peripheral writeValue:valData forCharacteristic:aChar type:CBCharacteristicWriteWithResponse];
            }
        }
    }
    
    if ([service.UUID isEqual:self.class.uartServiceUUID]) {    // Uart service (raw sensors).
        
        for (CBCharacteristic *c in [service characteristics]) {
            
            NSLog(@"Characteristic found with UUID: %@", c.UUID);
            
            if ([c.UUID isEqual:self.class.rxCharacteristicUUID]) {
                NSLog(@"Found RX characteristic");
                [peripheral setNotifyValue:YES forCharacteristic:c];
            } else if ([c.UUID isEqual:self.class.txCharacteristicUUID]) {
                self.txCharacteristic = c;
                NSLog(@"Found TX characteristic for ring1");
            }
            
        }
        
    }
    
}

/*
 Invoked upon completion of a -[readValueForCharacteristic:] request or on the reception of a notification/indication.
 */
- (void) peripheral:(CBPeripheral *)aPeripheral didUpdateValueForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error {
    
    if ([characteristic.UUID isEqual:self.class.rxCharacteristicUUID]) {
        
        rxLength = [characteristic value].length;
        memcpy(rxBytes, [[characteristic value] bytes], rxLength);
        //rxBytes = (unsigned char *)[[characteristic value] bytes];
        
        [[NSNotificationCenter defaultCenter] postNotificationName:@"updateValueForCharacteristicNotification" object:self userInfo:nil];
        
    }
    
}

- (void)peripheral:(CBPeripheral *)peripheral didWriteValueForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error {

    sendingData = false;
    
    if (error) {
        NSLog(@"write error");
        
        writeTrials++;
        if(writeTrials >= MAX_SEND_TRIALS) {
            NSLog(@"Cannot send packet!");
            [[NSNotificationCenter defaultCenter] postNotificationName:@"writeErrorNotification" object:self userInfo:nil];
        } else {
            [self writeBle: withResponseLast]; // Retry sending the last message.
        }
        return;
    }
    
    NSLog(@"write ok");
    [[NSNotificationCenter defaultCenter] postNotificationName:@"writeOkNotification" object:self userInfo:nil];
    
}

- (void)peripheral:(CBPeripheral *)peripheral didUpdateNotificationStateForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error
{
    NSLog(@"didUpdateNotificationStateForCharacteristic");
    
    if(error) {
        NSLog(@"error is: %@", [error localizedDescription]);
        //handle error
        //        BTcharacteristic=nil;
    } else {
        NSLog(@"Notification state updated correctly");
        [[NSNotificationCenter defaultCenter] postNotificationName:@"updateNotificationStateNotification" object:self userInfo:nil];
    }
    //[_peripheral readValueForCharacteristic:characteristic];
}

- (void) writeBle:(BOOL)withResponse {
    
    if(withResponse) {
        if(!sendingData) { // Send only one command at a time.
            sendingData = true;
            if ((self.txCharacteristic.properties & CBCharacteristicPropertyWrite) != 0) {
                [blePeripheral writeValue:bleBuff forCharacteristic:self.txCharacteristic type:CBCharacteristicWriteWithResponse];
            } else {
                NSLog(@"No write property on TX characteristic, %lu.", (unsigned long)txCharacteristic.properties);
            }
        }
    } else {
        if ((self.txCharacteristic.properties & CBCharacteristicPropertyWriteWithoutResponse) != 0) {
            [blePeripheral writeValue:bleBuff forCharacteristic:self.txCharacteristic type:CBCharacteristicWriteWithoutResponse];
        } else {
            NSLog(@"No write property on TX characteristic, %lu.", (unsigned long)txCharacteristic.properties);
        }
    }
    
}

- (void) writeString:(NSString *)string withResponse:(BOOL)value {
    writeTrials = 0;
    bleBuff = [NSData dataWithBytes:string.UTF8String length:string.length];
    [self writeBle: value];
    withResponseLast = value;
}

- (void) writeChar:(unsigned char)cmd withResponse:(BOOL)value {
    writeTrials = 0;
    bleBuff = [NSData dataWithBytes:&cmd length:sizeof(cmd)];
    [self writeBle: value];
    withResponseLast = value;
}

- (void) writeArray:(unsigned char*)msg withResponse:(BOOL)value {
    writeTrials = 0;
    bleBuff = [NSData dataWithBytes:msg length:MSG_SIZE];
    [self writeBle: value];
    withResponseLast = value;
}

- (int) getRxLength {
    return rxLength;
}

- (void) getRxBytes:(unsigned char*)bytes {
    memcpy(bytes, rxBytes, rxLength);
}


@end
